(*
 *  LDLITE, a program for viewing *.dat files.
 *  Copyright (C) 1998  Paul J. Gyugyi
 *
 *  Translated to Pascal (Delphi 2) by Anders Isaksson 1998
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

unit ldliteVR_main;

interface

uses
  ldliteVRh, debug
  ;

procedure ldlite_parse(filename: string; Scale, RotX, RotY, RotZ: MyReal;
  vp: PVector3D; mp: PMatrix3D);

function SetMatrix(const a, b, c, d, e, f, g, h, i: MyReal): PMatrix3D;
function SetTransl(const x, y, z: MyReal): PVector3D;

implementation

uses
  SysUtils, Classes,
  ytab, stub
  ;

var
  part_name         : string;
  flip_scale        : vector3d;
  yyin              : TStringList;
  GlobMatrx         : Matrix3D;
  GlobVect          : Vector3D;

function SetMatrix(const a, b, c, d, e, f, g, h, i: MyReal): PMatrix3D;
begin
  GlobMatrx.a := a;
  GlobMatrx.b := b;
  GlobMatrx.c := c;
  GlobMatrx.d := d;
  GlobMatrx.e := e;
  GlobMatrx.f := f;
  GlobMatrx.g := g;
  GlobMatrx.h := h;
  GlobMatrx.i := i;
  Result := @GlobMatrx;
end;

function SetTransl(const x, y, z: MyReal): PVector3D;
begin
  GlobVect.x := x;
  GlobVect.y := y;
  GlobVect.z := z;
  Result := @GlobVect;
end;

(*
 * m3 = m2 * m1;
 *)

procedure transform_multiply(var t1: Pvector3d; var m1: Pmatrix3d;
  var t2: Pvector3d; var m2: Pmatrix3d;
  var t3: Pvector3d; var m3: Pmatrix3d);
begin
  if Assigned(t3) then FreeMem(t3);
  if Assigned(m3) then FreeMem(m3);
  t3 := AllocMem(sizeof(Vector3d));
  m3 := AllocMem(sizeof(Matrix3d));
  m3.a := m2.a * m1.a + m2.b * m1.d + m2.c * m1.g;
  m3.b := m2.a * m1.b + m2.b * m1.e + m2.c * m1.h;
  m3.c := m2.a * m1.c + m2.b * m1.f + m2.c * m1.i;
  m3.d := m2.d * m1.a + m2.e * m1.d + m2.f * m1.g;
  m3.e := m2.d * m1.b + m2.e * m1.e + m2.f * m1.h;
  m3.f := m2.d * m1.c + m2.e * m1.f + m2.f * m1.i;
  m3.g := m2.g * m1.a + m2.h * m1.d + m2.i * m1.g;
  m3.h := m2.g * m1.b + m2.h * m1.e + m2.i * m1.h;
  m3.i := m2.g * m1.c + m2.h * m1.f + m2.i * m1.i;
  t3.x := m2.a * t1.x + m2.b * t1.y + m2.c * t1.z + t2.x;
  t3.y := m2.d * t1.x + m2.e * t1.y + m2.f * t1.z + t2.y;
  t3.z := m2.g * t1.x + m2.h * t1.y + m2.i * t1.z + t2.z;
end;

procedure ldlite_parse(filename: string; Scale, RotX, RotY, RotZ: MyReal;
  vp: PVector3D; mp: PMatrix3D);
var
  i                 : integer;
  fp                : TextFile;
  v1                : PVector3d;
  m1                : PMatrix3d;
  v2                : PVector3d;
  m2                : PMatrix3d;
  v3                : PVector3d;
  m3                : PMatrix3d;
  v4                : PVector3d;
  m4                : PMatrix3d;
  v5                : PVector3d;
  m5                : PMatrix3d;
  DoFree            : boolean;

begin
  if (vp = nil) and (mp = nil) then begin
    DoFree := True;
    v1 := savevec(0.0, 0.0, 0.0);
    m1 := savemat(
      1.0, 0.0, 0.0,
      0.0, Cos(RotX), Sin(RotX),
      0.0, -Sin(RotX), Cos(RotX));
    v2 := savevec(0.0, 0.0, 0.0);
    m2 := savemat(
      Cos(RotY), 0.0, Sin(RotY),
      0.0, 1.0, 0.0,
      -Sin(RotY), 0.0, Cos(RotY));
    v3 := nil;
    m3 := nil;
    transform_multiply(v1, m1, v2, m2, v3, m3);
    v4 := savevec((zGetRowsize() / 2), (zGetColsize() / 2), 0.0);
    (*
      ldraw_commandline_opts.A.a=1.0;
      ldraw_commandline_opts.A.b=0.0;
      ldraw_commandline_opts.A.c=1.0;
      ldraw_commandline_opts.A.d=0.5;
      ldraw_commandline_opts.A.e=1.0;
      ldraw_commandline_opts.A.f=-0.5;
      ldraw_commandline_opts.A.g=-1.0;
      ldraw_commandline_opts.A.h=0.0;
      ldraw_commandline_opts.A.i=1.0;
    *)
    m4 := savemat(
      Cos(RotZ), Sin(RotZ), 0.0,
      -Sin(RotZ), Cos(RotZ), 0.0,
      0.0, 0.0, 1.0);
    v5 := nil;
    m5 := nil;
    transform_multiply(v3, m3, v4, m4, v5, m5);
    (*
      m5 := savemat(
          1.0,  0.0,  1.0),
          0.5,  1.0, -0.5),
         -1.0,  0.0,  1.0)
         );
    *)
  end
  else begin
    DoFree := False;
    v5 := vp;
    m5 := mp;
  end;

  m5.a := m5.a * Scale;
  m5.b := m5.b * Scale;
  m5.c := m5.c * Scale;
  m5.d := m5.d * Scale;
  m5.e := m5.e * Scale;
  m5.f := m5.f * Scale;
  m5.g := m5.g * Scale;
  m5.h := m5.h * Scale;
  m5.i := m5.i * Scale;
  (*
    debug.print(format('%f %f %f', [m5.a, m5.b, m5.c]));
    debug.print(format('%f %f %f', [m5.d, m5.e, m5.f]));
    debug.print(format('%f %f %f', [m5.g, m5.h, m5.i]));
    debug.print('');
  *)

  include_stack_ptr := 0;
  current_color[include_stack_ptr] := 7;
  current_translation[include_stack_ptr] := v5^;
  current_transform[include_stack_ptr] := m5^;

  yyin := TStringList.Create;
  try
    yyin.LoadFromFile(filename);
    yyparse(yyin, False);
  finally
    yyin.Free;
    if DoFree then begin
      FreeMem(v1);
      FreeMem(v2);
      FreeMem(v3);
      FreeMem(v4);
      FreeMem(v5);
      FreeMem(m1);
      FreeMem(m2);
      FreeMem(m3);
      FreeMem(m4);
      FreeMem(m5);
    end
  end
end;

end.

