{*
 *  LDLITE, a program for viewing *.dat files.
 *  Copyright (C) 1998  Paul J. Gyugyi
 *
 *  Translated to Pascal (Delphi 2) by Anders Isaksson 1998
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*}

unit ytab;

interface

uses
  Classes,
  ldliteVRh
  ;

var
  current_color:        array[0..MAX_INCLUDE_DEPTH] of integer;
  current_translation:  array[0..MAX_INCLUDE_DEPTH] of Vector3d;
  current_transform:    array[0..MAX_INCLUDE_DEPTH] of Matrix3d;
  include_stack_ptr:    integer = 0;
  LDrawDir:             string  = 'c:\ldraw';
  CachePrimitives:      Boolean = True;
  CacheParts:           Boolean = False;
  BCPolCnt:             integer;
  BCStdCnt:             integer;
  BCNoSave:             integer;

function  savevec(const x, y, z: MyReal): PVector3d;
function  savemat(const a, b, c, d, e, f, g, h, i: MyReal): PMatrix3d;
function  yyparse(const md: TStringList; Primary: Boolean):  integer;

procedure StartNewPiece;
procedure SaveAsBlockCAD(pName: string);

implementation

uses
  Windows, SysUtils, Dialogs,
  lexyy, stub, Main, Debug
  ;

type
  TCoord =   array[0..2] of integer;
  TPoly  =   record
    NumVertex:    integer;
    col:          integer;
    VertexList:   array[0..3] of TCoord;
  end;
  TPolys =   array[0..1023] of TPoly;
  TStuds =   array[0..511] of TCoord;

var
  Cache:      TStringList;
  BCPolys:  TPolys;
  BCStuds:  TStuds;

procedure StartNewPiece;
begin
  BCPolCnt := 0;
  BCStdCnt := 0;
  BCNoSave := 0;
end;

procedure SaveAsBlockCAD(pName: string);
var
  f:        TStringList;
  s:        string;
  i, j:     integer;
  L0,B0,H0,
  L9,B9,H9: integer;
begin
  if BCPolCnt > 0 then begin
    f := TStringList.Create;
    try
      L0 := 99999;
      B0 := 99999;
      H0 := 99999;
      L9 := -99999;
      B9 := -99999;
      H9 := -99999;
      for i := 0 to Pred(BCPolCnt) do begin
        for j := 0 to Pred(BCPolys[i].NumVertex) do begin
          if BCPolys[i].VertexList[j, 0] > L9 then
            L9 := BCPolys[i].VertexList[j, 0];
          if BCPolys[i].VertexList[j, 1] > B9 then
            B9 := BCPolys[i].VertexList[j, 1];
          if BCPolys[i].VertexList[j, 2] > H9 then
            H9 := BCPolys[i].VertexList[j, 2];
          if BCPolys[i].VertexList[j, 0] < L0 then
            L0 := BCPolys[i].VertexList[j, 0];
          if BCPolys[i].VertexList[j, 1] < B0 then
            B0 := BCPolys[i].VertexList[j, 1];
          if BCPolys[i].VertexList[j, 2] < H0 then
            H0 := BCPolys[i].VertexList[j, 2];
        end
      end;
      { Write polygons }
      f.Add(Format('[%s]', [pName]));
      f.Add(Format('NP=%d', [BCPolCnt]));
      f.Add(Format('BB=%d,%d,%d', [B9-B0, L9-L0, H9-H0]));
//      if (LDrawPatch <> 0) then
//        f.Add(Format('LD=%d', [LDrawPatch]));
      for i := 0 to Pred(BCPolCnt) do begin
        s := Format('P%d=%d, %d', [i+1, BCPolys[i].NumVertex, 1 {Border}]);
        for j := 0 to Pred(BCPolys[i].NumVertex) do begin
          s := s + Format(', %d,%d,%d',
                  [BCPolys[i].VertexList[j, 1] - B0,
                   BCPolys[i].VertexList[j, 0] - L0,
                   BCPolys[i].VertexList[j, 2] - H0]);
        end;
      { Write detail level - optional }
        if {(BCPolys[i].DetLev > 0) or} (BCPolys[i].col <> 7) then
          s := s + Format(', %d', [0 {BCPolys[i].DetLev}]);
      { Write fixed color - optional }
        if (BCPolys[i].col <> 7) then
          s := s + Format(', -1, %d', [BCPolys[i].col]);
        f.Add(s);
      end;
      { Write bitmappoly/file - optional }
//      if (fBitMapPoly > -1) then begin
//        f.Add(Format('BM=%d, %s', [Succ(fBitMapPoly), fBitMapFile]));
//      end;
      { Write drawing order per octant 'Qx' }
      { needs sorting in BlockCAD           }
      for i := 0 to 7 do begin
        s := Format('Q%d=%d ', [Succ(i), BCPolCnt]);
        for j := 0 to Pred(BCPolCnt) do begin
          s := s + Format(',%d', [j+1]);
        end;
        f.Add(s);
      end;
      { Write topstuds info 'TS' }
      s := Format('TS=%d,%d', [0, 0]);
//      for i := 0 to Pred(TopStudCount) do begin
//        s := s + Format(', %d,%d', [TopStudPos[i].x, TopStudPos[i].y]);
//        if (TopStudPos[i].z <> Polys[TopStudPoly].VertexList[0].z) then
//          s := s + Format(';%d', [TopStudPos[i].z]);
//      end;
      f.Add(s);
      { Write bottomstuds info 'BS' }
      s := Format('BS=%d,%d', [0, 0]);
//      for i := 0 to Pred(BtmStudCount) do begin
//        s := s + Format(', %d,%d', [BtmStudPos[i].x, BtmStudPos[i].y]);
//        if (BtmStudPos[i].z <> Polys[BtmStudPoly].VertexList[0].z) then
//          s := s + Format(';%d', [BtmStudPos[i].z]);
//      end;
      f.Add(s);
      { Write rotation offsets 'RO' - optional}
//      for i := 0 to 3 do begin
//        if (RotOffset[i, 0] <> 0) or (RotOffset[i, 1] <> 0) then begin
//          s := Format('RO=%d,%d, %d,%d, %d,%d, %d,%d',
//                      [RotOffset[0, 0], RotOffset[0, 1],
//                       RotOffset[1, 0], RotOffset[1, 1],
//                       RotOffset[2, 0], RotOffset[2, 1],
//                       RotOffset[3, 0], RotOffset[3, 1]
//                      ]);
//          f.Add(s);
//          Break;
//        end
//      end;
      f.SaveToFile('C:\Temp\ld2bc.txt');
    finally
      f.Clear;
    end
  end
end;

function savevec(const x, y, z: MyReal): PVector3d;
begin
  Result := AllocMem(sizeof(vector3d));
  Result.x := x;
  Result.y := y;
  Result.z := z;
end;

function savemat(const a, b, c, d, e, f, g, h, i: MyReal): PMatrix3d;
begin
  Result := AllocMem(sizeof(matrix3d));
  Result.a := a;
  Result.b := b;
  Result.c := c;
  Result.d := d;
  Result.e := e;
  Result.f := f;
  Result.g := g;
  Result.h := h;
  Result.i := i;
end;

procedure transform_vec_inplace(var v: vector3d);
var
 tmp: vector3d;
 m:   pmatrix3d;
 t:   pvector3d;
begin
 m := @current_transform[include_stack_ptr];
 t := @current_translation[include_stack_ptr];
 tmp.x := m.a * v.x + m.b * v.y + m.c * v.z + t.x;
 tmp.y := m.d * v.x + m.e * v.y + m.f * v.z + t.y;
 tmp.z := m.g * v.x + m.h * v.y + m.i * v.z + t.z;
 v := tmp;
end;

function push_transform(var t: Vector3d; var m: Matrix3d): integer;
var
  oldt: PVector3d;
  newt: PVector3d;
  oldm: PMatrix3d;
  newm: PMatrix3d;
begin

  if (include_stack_ptr <= 0) then begin
	  ShowMessage('Error in transform stack');
	  Halt;
  end;
  oldm := @current_transform[include_stack_ptr - 1];
  oldt := @current_translation[include_stack_ptr - 1];
  newm := @current_transform[include_stack_ptr];
  newt := @current_translation[include_stack_ptr];

  newm.a := oldm.a * m.a + oldm.b * m.d + oldm.c * m.g;
  newm.b := oldm.a * m.b + oldm.b * m.e + oldm.c * m.h;
  newm.c := oldm.a * m.c + oldm.b * m.f + oldm.c * m.i;
  newm.d := oldm.d * m.a + oldm.e * m.d + oldm.f * m.g;
  newm.e := oldm.d * m.b + oldm.e * m.e + oldm.f * m.h;
  newm.f := oldm.d * m.c + oldm.e * m.f + oldm.f * m.i;
  newm.g := oldm.g * m.a + oldm.h * m.d + oldm.i * m.g;
  newm.h := oldm.g * m.b + oldm.h * m.e + oldm.i * m.h;
  newm.i := oldm.g * m.c + oldm.h * m.f + oldm.i * m.i;
  newt.x := oldm.a * t.x + oldm.b * t.y + oldm.c * t.z + oldt.x;
  newt.y := oldm.d * t.x + oldm.e * t.y + oldm.f * t.z + oldt.y;
  newt.z := oldm.g * t.x + oldm.h * t.y + oldm.i * t.z + oldt.z;
  m := newm^;
  t := newt^;
//  current_transform[include_stack_ptr] := m;
//  current_translation[include_stack_ptr] := t;
  Result := 0;
end;

type
  PLDrawLine = ^TLDrawLine;
  TLDrawLine = record  {1 colour x y z a b c d e f g h i part.dat}
    Typ, Color: integer;
    f:          array[2..13] of MyReal;
    s:          string[255];
  end;

var
  Level:    integer = -1;
  sl:       TLdrawLine;

function GetColor: integer;
begin
  Result := sl.Color;
  if (Result = 16) then
    Result := current_color[include_stack_ptr]
  else if (Result = 24) then
	  Result := edge_color(current_color[include_stack_ptr])
end;

function MakeVect(const start: integer): Vector3d;
begin
  Result.x := sl.f[start];
  Result.y := sl.f[start+1];
  Result.z := sl.f[start+2];
end;

function MakeMat(const start: integer): Matrix3d;
begin
  Result.a := sl.f[start];
  Result.b := sl.f[start+1];
  Result.c := sl.f[start+2];
  Result.d := sl.f[start+3];
  Result.e := sl.f[start+4];
  Result.f := sl.f[start+5];
  Result.g := sl.f[start+6];
  Result.h := sl.f[start+7];
  Result.i := sl.f[start+8];
end;

procedure Split(const q: string);
var
  c:  integer;
  p:  PChar;
  s:  string;

  function ScanInt: integer;
  var m: integer;
  begin
    m := 1;
    Result := 0;
    while (p^ = ' ') or (p^ = #9) do begin
      Inc(p);
    end;
    if (p^ = '-') then begin
      m := -1;
      Inc(p);
    end;
    while (p^ >= '0') and (p^ <= '9') do begin
      Result := 10*Result + Ord(p^) - Ord('0');
      Inc(p);
    end;
    Result := Result * m;
  end;

  function ScanFloat: MyReal;
  var m,d: MyReal;
  begin
    m := 1.0;
    Result := 0;
    while (p^ = ' ') or (p^ = #9) do begin
      Inc(p);
    end;
    if (p^ = '-') then begin
      m := -1.0;
      Inc(p);
    end;
    while (p^ >= '0') and (p^ <= '9') do begin
      Result := 10*Result + Ord(p^) - Ord('0');
      Inc(p);
    end;
    if (p^ = '.') then begin
      d := 0.1;
      Inc(p);
      while (p^ >= '0') and (p^ <= '9') do begin
        Result := Result + (Ord(p^) - Ord('0'))*d;
        Inc(p);
        d := d * 0.1;
      end;
    end;
    Result := Result * m;
  end;

begin
  c := 0;
  sl.Typ := -1;
  s := q + #0;
  if (Length(s) > 0) then begin
    p := @s[1];
    sl.Typ := ScanInt;
    if sl.Typ = 0 then begin
      sl.s := p;
    end
    else begin
      sl.Color := ScanInt;
      c := Low(sl.f);
      while (c <= High(sl.f)) and (p^ <> #0) do begin
        sl.f[c] := ScanFloat;
        Inc(c);
      end;
      if sl.Typ = 1 then begin
        while (p^ = ' ') or (p^ = #9) do
          Inc(p);
        sl.s := p;
      end
    end;
  end;
  s := '';
end;

function yyparse(const md: TStringList; Primary: boolean): integer;
var
  ix:       integer;
  tl:       TStringList;
  Typ:      integer;
  Step:     integer;
  cl:       integer;
  v1, v2, v3, v4: Vector3d;
  m1:       Matrix3d;
  filename: string;
  BaseName: string;
  Found:    boolean;
  i:        integer;
  P:        Boolean;
  PL:       PLDrawLine;
  x1, y1, z1: MyReal;
  x2, y2, z2: MyReal;
  d1, d2:   MyReal;
begin
  Inc(Level);
  Step := 0;
  try
    for ix := 0 to Pred(md.Count) do begin
      if Length(md[ix]) = 0 then
        continue;
//      Debug.Print(md[ix]);
      if (md.Objects[ix] = nil) then begin
        Split(md[ix]);
        if Primary then begin
          New(PL);
          PL^ := sl;
          md.Objects[ix] := TObject(PL);
        end
      end
      else begin
        sl := PLDrawLine(md.Objects[ix])^;
      end;
      case sl.Typ of
      0:  {comment or meta-command}
        begin
          if (CompareText(sl.s, 'step') = 0) then begin
            Inc(Step);
            zStep(Step, 1);
          end
          else begin
            if (Level = 0) then
              platform_comment(sl.s, Level);
          end
        end;
      1:  {Part-file reference}
          {1 colour x y z a b c d e f g h i part.dat}
        begin
          cl := GetColor;
          v1 := MakeVect(2);
          m1 := MakeMat(5);
          BaseName := sl.s;
          if Pos('STUD', UpperCase(BaseName)) > 0 then begin
            Inc(BCNoSave);
            if BCStdCnt < High(BCStuds) then begin
              BCStuds[BCStdCnt][0] := Round(v1.x);
              BCStuds[BCStdCnt][1] := Round(v1.y);
              BCStuds[BCStdCnt][2] := Round(v1.z);
              Inc(BCStdCnt);
            end;
          end;
          Found := False;
          P := False;
          if Assigned(Cache) then begin
            i := Cache.IndexOf(BaseName);
            if (i >= 0) then begin
              tl := TStringList(Cache.Objects[i]);
              Found := True;
            end
          end;

          if not Found then begin
            FileName := Format('%s\p\%s',[ldrawdir, BaseName]);
            if not FileExists(FileName) then begin
              FileName := Format('%s\parts\%s',[ldrawdir, BaseName]);
              if not FileExists(FileName) then begin
                FileName := Format('%s\models\%s',[ldrawdir, BaseName]);
                if not FileExists(FileName) then begin
                  FileName := Format('%s',[ldrawdir, BaseName]);
                  Found := FileExists(FileName);
                end
                else Found := True;
              end
              else begin
                Found := True;
                P := CacheParts;
              end
            end
            else begin
              Found := True;
              P := CachePrimitives;
            end;

            if Found then begin
//              Debug.Print(FileName);
              tl := TStringList.Create;
              tl.LoadFromFile(FileName);
              Cache.AddObject(BaseName, tl);
            end;
          end;

          if Found then begin
            Inc(include_stack_ptr);
      		  push_transform(v1, m1);
	  	      current_color[include_stack_ptr] := cl;
            yyparse(tl, P);
            tl := nil;
            Dec(include_stack_ptr);
            if include_stack_ptr = 0 then zStep(Step, 0);
          end;
          if Pos('STUD', UpperCase(BaseName)) > 0 then begin
            Dec(BCNoSave);
          end
        end;
      2:  {Line}
          {2 colour x1 y1 z1 x2 y2 z2}
        begin
          cl := GetColor;
          v1 := MakeVect(2);
          v2 := MakeVect(5);
        	transform_vec_inplace(v1);
        	transform_vec_inplace(v2);
        	render_line(v1, v2, cl);
          { later: fix borders on BC-polys }
        end;
      3:  {Triangle}
          {3 colour x1 y1 z1 x2 y2 z2 x3 y3 z3}
        begin
          cl := GetColor;
          v1 := MakeVect(2);
          v2 := MakeVect(5);
          v3 := MakeVect(8);
        	transform_vec_inplace(v1);
          transform_vec_inplace(v2);
          transform_vec_inplace(v3);
          render_triangle(v1, v2, v3, cl);
          if BCNoSave = 0 then begin
            if BCPolCnt < High(BCPolys) then begin
              with BCPolys[BCPolCnt] do begin
                NumVertex := 3;
                col := cl;
                VertexList[0, 0] := Round(v1.x);
                VertexList[0, 1] := Round(v1.y);
                VertexList[0, 2] := Round(v1.z);
                VertexList[1, 0] := Round(v2.x);
                VertexList[1, 1] := Round(v2.y);
                VertexList[1, 2] := Round(v2.z);
                VertexList[2, 0] := Round(v3.x);
                VertexList[2, 1] := Round(v3.y);
                VertexList[2, 2] := Round(v3.z);
              end;
              Inc(BCPolCnt);
            end
          end
        end;
      4:  {Quadrilateral}
          {4 colour x1 y1 z1 x2 y2 z2 x3 y3 z3 x4 y4 z4}
        begin
          cl := GetColor;
          v1 := MakeVect(2);
          v2 := MakeVect(5);
          v3 := MakeVect(8);
          v4 := MakeVect(11);
        	transform_vec_inplace(v1);
          transform_vec_inplace(v2);
          transform_vec_inplace(v3);
          transform_vec_inplace(v4);
        	render_quad(v1, v2, v3, v4, cl);
          if BCNoSave = 0 then begin
            if BCPolCnt < High(BCPolys) then begin
              with  BCPolys[BCPolCnt] do begin
                NumVertex := 4;
                col := cl;
                x1 := (v2.x - v1.x);
                y1 := (v2.y - v1.y);
                z1 := (v2.z - v1.z);
                x2 := (v4.x - v3.x);
                y2 := (v4.y - v3.y);
                z2 := (v4.z - v3.z);
                d1 := Sqrt(Sqr(x1) + Sqr(y1) + Sqr(z1));
                d2 := Sqrt(Sqr(x2-x1) + Sqr(y2-y1) + Sqr(z2-z1));
                if d2 < d1 then begin { Bowtie }
                  VertexList[0, 0] := Round(v1.x);
                  VertexList[0, 1] := Round(v1.y);
                  VertexList[0, 2] := Round(v1.z);
                  VertexList[1, 0] := Round(v2.x);
                  VertexList[1, 1] := Round(v2.y);
                  VertexList[1, 2] := Round(v2.z);
                  VertexList[3, 0] := Round(v3.x);
                  VertexList[3, 1] := Round(v3.y);
                  VertexList[3, 2] := Round(v3.z);
                  VertexList[2, 0] := Round(v4.x);
                  VertexList[2, 1] := Round(v4.y);
                  VertexList[2, 2] := Round(v4.z);
                end
                else begin
                  VertexList[0, 0] := Round(v1.x);
                  VertexList[0, 1] := Round(v1.y);
                  VertexList[0, 2] := Round(v1.z);
                  VertexList[1, 0] := Round(v2.x);
                  VertexList[1, 1] := Round(v2.y);
                  VertexList[1, 2] := Round(v2.z);
                  VertexList[2, 0] := Round(v3.x);
                  VertexList[2, 1] := Round(v3.y);
                  VertexList[2, 2] := Round(v3.z);
                  VertexList[3, 0] := Round(v4.x);
                  VertexList[3, 1] := Round(v4.y);
                  VertexList[3, 2] := Round(v4.z);
                end
              end;
              Inc(BCPolCnt);
            end
          end
        end;
      5:  {Optional line}
          {5 colour x1 y1 z1 x2 y2 z2 x3 y3 z3 x4 y4 z4}
        begin
          cl := GetColor;
          v1 := MakeVect(2);
          v2 := MakeVect(5);
          v3 := MakeVect(8);
          v4 := MakeVect(11);
        	transform_vec_inplace(v1);
          transform_vec_inplace(v2);
          transform_vec_inplace(v3);
          transform_vec_inplace(v4);
        	render_five(v1, v2, v3, v4, cl);
          { does not exist in BC }
        end;
      else {something else}
        begin
//          Debug.Print('Error' + md[ix]);
        end;
      end {case}
    end;
  finally
    Dec(Level);
  end;
end;

procedure ClearCache;
var i, j: integer;
begin
  for i := 0 to Pred(Cache.count) do begin
    with TStringList(Cache.Objects[i]) do begin
      for j := 0 to Pred(Count) do begin
        if Assigned(Objects[j]) then
          Dispose(Pointer(Objects[j]));
      end;
      Free;
    end
  end;
  Cache.Free;
end;

initialization
  Cache := TStringList.Create;
  Cache.Sorted := True;
finalization
  ClearCache;
end.
